/*
    fsresize 0.08 (liposuck)
    Copyright (C) 1998-1999  Andrew Clausen  <clausen@alphalink.com.au>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can also be contacted at:

    Andrew Clausen
    18 Shaw St
    Ashwood, 3147
    Victoria, Australia

    Some of this code is from mkdosfs (Dave Hudson dave@humbug.demon.co.uk)
*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "libresize/libresize.h"

msdos_filesystem   filesystem;

int     display_info = 0;
int     backup = 0;
int     restore = 0;
int     resize = 0;
FILE   *backup_handle;

void standard_error () {
    perror("fsresize");
    exit (1);
}

void error (char *msg) {
    fprintf(stderr, "fsresize: %s\n", msg);
    exit (1);
}

void usage () {
    printf (
"fsresize 0.08 (liposuck), Copyright (C) 1998-1999 Andrew Clausen\n"
"fsresize comes with ABSOLUTELY NO WARRANTY; it is possible that this program\n"
"will destroy your data.  Read the COPYING file for details.\n\n"
    );

    printf (
"Usage:  fsresize  [<partition> <new-size>[km]] [<partition> -i]\n"
"        fsresize  <partition> --backup <backup-file>\n"
"        fsresize  <partition> --restore <backup-file>\n"
"   <new-size> is the size to make <partition>, in megabytes, kilobytes or\n"
"sectors. (can be followed by M, K, or S; K by default)\n"
"\n"
"  -b, --backup <backup-file>    backup important info to <backup-file>\n"
"  -h, --help                    display this help and exit\n"
"  -i, --info                    display information about <device> and exit\n"
"  -r, --restore <backup-file>   restore from <backup-file>, (if it died!)\n"
"\n"
"examples: \n"
"    fsresize /dev/hda9 -i                    prints info about /dev/hda9\n"
"    fsresize /dev/hda2 738M                  resizes /dev/hda2 to 738 megs\n"
"    fsresize /dev/sdb2 --backup /fd0/backup  backs up /dev/sdb2\n"
"    fsresize /dev/sdb2 234M --backup /fd0/backup  back up, then resize\n");
    exit(1);
}

int
main (int argc, char *argv[]) {
    int     opt_status;
    char   *device_name;
    char    partition;
    int     i;
    long long new_size;

    struct option long_options[] =
    {
        {"help", 0, 0, 'h'},
        {"info", 0, 0, 'i'},
        {"backup", 1, 0, 'b'},
        {"restore", 1, 0, 'r'},
        {0, 0, 0, 0}
    };

    while (1) {
        opt_status = getopt_long (argc, argv, "?b:hir:", long_options, NULL);
        if (opt_status == EOF) break;

        switch (opt_status) {
        case 'b':
            backup = 1;
            backup_handle = fopen (optarg, "w");
            if (backup_handle == NULL) standard_error ();
            break;
            
        case 'i':
            display_info = 1;
            break;

        case 'h':
        case '?':
            usage ();
            break;

        case 'r':
            restore = 1;
            backup_handle = fopen (optarg, "r");
            if (backup_handle == NULL) standard_error ();
            break;
        }
    }

    if (backup + restore + display_info > 1)
        error ("only one command of -b, -i and -r allowed");

    if (backup + restore + display_info == 0) resize = 1;

    if (argv[optind] == NULL) usage();

    device_name = (char *) strdup (argv[optind]);
    device_name[strlen(device_name) - 1] = 0;
    partition = argv[optind][strlen (argv[optind]) - 1] - '0';

    if (resize || argv[optind+1] != NULL) {
        resize = 1;
        if (argv [optind+1] == NULL)
            error ("you need to specify a size or command");

        new_size = strtol (argv [optind+1], (char **)NULL, 10);

        /* I don't believe anyone would use > 32G for *DOS* */
        if (new_size > 32 * 1024 * 1024)
            error ("You *DO* realise that that the new size is in K, hmmm?");

        switch (argv [optind+1][strlen (argv [optind+1]) - 1]) {
        case 'm':
        case 'M':
            new_size *= 1024 * 1024;
            break;

        case 's':
        case 'S':
            new_size *= 512;
            break;

        default:
            new_size *= 1024;
            break;
        }
    }

    if (!trimlib_init_filesystem (&filesystem, device_name, partition))
        exit(1);

    if (display_info) trimlib_print_info (&filesystem);

    if (backup)
        if (!trimlib_create_backup (backup_handle, &filesystem))
            exit(1);

    if (restore)
        if (!trimlib_restore_backup (backup_handle, &filesystem))
            exit(1);

    if (resize)
        if (!trimlib_resize_filesystem (&filesystem, new_size))
            exit(1);

    if (backup || restore) fclose (backup_handle);

    trimlib_close_filesystem (&filesystem);
}
