/*
    libresize 0.08
    Copyright (C) 1998-1999  Andrew Clausen  <clausen@alphalink.com.au>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can also be contacted at:

    Andrew Clausen
    18 Shaw St
    Ashwood, 3147
    Victoria, Australia

*/

#include "libresize.h"
#include "../libfdisk/libfdisk.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

static void
backup_error (msdos_filesystem *fs) {
    trimlib_error (fs, "Error writing backup - %s", strerror (errno));
}

static int
backup_header (FILE *stream, msdos_filesystem *fs) {
    if (fprintf (stream, "Liposuck Backup\n\n") == 0) {
        backup_error (fs);
        return 0;
    }
    return 1;
}

static int
backup_partition_size (FILE *stream, msdos_filesystem *fs) {
    if (fprintf (stream, "Partition size: %lld bytes\n", fs->partition_size)
          == 0) {
        backup_error (fs);
        return 0;
    }
    return 1;
}

static int
backup_filesystem_type (FILE *stream, msdos_filesystem *fs) {
    int status;

    if (fprintf (stream, "Filesystem type: ") == 0) {
        backup_error (fs);
        return 0;
    }

    switch (fs->fs_type) {
        case FAT12: status = fprintf (stream, "FAT12\n"); break;
        case FAT16: status = fprintf (stream, "FAT16\n"); break;
        case FAT32: status = fprintf (stream, "FAT32\n"); break;
    }

    if (status == 0) {
        backup_error (fs);
        return 0;
    }

    return 1;
}

static int
backup_boot_sector (FILE *stream, msdos_filesystem *fs) {
    if (fprintf (stream, "\nBoot sector:\n") == 0) {
        backup_error (fs);
        return 0;
    }

    if (fwrite (&fs->boot_sector, 512, 1, stream) != 1) {
        backup_error (fs);
        return 0;
    }

    return 1;
}

static int
backup_fat (FILE *stream, msdos_filesystem *fs) {
    if (fprintf (stream, "\n\nFAT:\n") == 0) {
        backup_error (fs);
        return 0;
    }

    if (fwrite (fs->fats[0], fs->fat_size, 1, stream) != 1) {
        backup_error (fs);
        return 0;
    }

    return 1;
}

/*
    stores:
    * entire root directory for FAT16
    * first root directory cluster for FAT32
*/
static int
backup_root_dir (FILE *stream, msdos_filesystem *fs) {
    int count;
    int sector;

    if (fprintf (stream, "\n\nRoot directory:\n") == 0) {
        backup_error (fs);
        return 0;
    }

    if (fs->fs_type == FAT16) {
        count = fs->root_dir_entry_count * sizeof (directory_entry)
                    / fs->sector_size;

        for (sector = 0; sector < count; sector++)  {
            if (!trimlib_seek_read (fs, fs->buffer,
                fs->root_offset + sector * 512, 512)) return 0;
                
            if (fwrite (fs->buffer, 512, 1, stream) != 1) {
                backup_error (fs);
                return 0;
            }
        }
    } else if (fs->fs_type == FAT32) {
        if (!trimlib_read_cluster (fs, fs->buffer, fs->root_cluster)) return 0;
        if (fwrite (fs->buffer, fs->cluster_size, 1, stream) != 1)
            return 0;
    }

    return 1;
}

static int
backup_footer (FILE *stream, msdos_filesystem *fs) {
    if (fprintf (stream, "\n\nend of Liposuck backup\n") == 0) {
        backup_error (fs);
        return 0;
    }

    return 1;
}

/*
    Creates a backup of fs, storing it in stream
*/
int
trimlib_create_backup (FILE *stream, msdos_filesystem *fs) {
    if (!trimlib_prepare_filesystem (fs)) return 0;

    if (!backup_header (stream, fs)) return 0;
    if (!backup_partition_size (stream, fs)) return 0;
    if (!backup_filesystem_type (stream, fs)) return 0;
    if (!backup_boot_sector (stream, fs)) return 0;
    if (!backup_fat (stream, fs)) return 0;
    if (!backup_root_dir (stream, fs)) return 0;
    if (!backup_footer (stream, fs)) return 0;

    trimlib_unprepare_filesystem (fs);
    return 1;
}

static int
read_heading (FILE *stream, msdos_filesystem *fs, char *heading) {
    char    buf[128];

    if (fscanf (stream, "\n\n%[^:]:\n", buf) == 0) {
        trimlib_error (fs, "Error reading backup at section '%s'", heading);
        return 0;
    }

    if (strcmp (heading, buf) != 0) {
        trimlib_error (fs, "Invalid backup file - expected section '%s'\n"
                           "but found '%s'", heading, buf);
        return 0;
    }

    return 1;
}

static int 
count_free_clusters (msdos_filesystem *fs) {
    cluster_t   cluster;

    fs->free_cluster_count = 0;
    for (cluster=2; cluster<fs->cluster_count+2; cluster++)
        if (trimlib_is_available (fs, cluster)) fs->free_cluster_count++;

    return 1;
}

static void
restore_error (msdos_filesystem *fs) {
    trimlib_error (fs, "Error reading from backup - %s", strerror (errno));
}

static int
read_header (FILE *stream, msdos_filesystem *fs) {
    char    buffer[10];

    if (fscanf (stream, "%9s Backup\n\n", buffer) == 0) {
        restore_error (fs);
        return 0;
    }

    if (strncmp (buffer, "Liposuck", 9) != 0) {
        trimlib_error (fs, "Invalid backup file");
        return 0;
    }

    return 1;
}

static int
read_partition_size (FILE *stream, msdos_filesystem *fs) {
    if (fscanf (stream, "Partition size: %lld bytes\n", &fs->partition_size)
           == 0) {
        restore_error (fs);
        return 0;
    }
    return 1;
}

static int
read_filesystem_type (FILE *stream, msdos_filesystem *fs) {
    if (fscanf (stream, "Filesystem type: %s\n", fs->buffer) == 0) {
        restore_error (fs);
        return 0;
    }

    if (strcmp (fs->buffer, "FAT12") == 0) fs->fs_type = FAT12;
    else if (strcmp (fs->buffer, "FAT16") == 0) fs->fs_type = FAT16;
    else if (strcmp (fs->buffer, "FAT32") == 0) fs->fs_type = FAT32;
    else {
        trimlib_error (fs, "trimlib_restore_backup:  unknown filesystem: %s",
                       fs->buffer);
        return 0;
    }

    return 1;
}

static int
read_boot_sector (FILE *stream, msdos_filesystem *fs) {
    if (!read_heading (stream, fs, "Boot sector")) return 0;

    if (fread (&fs->boot_sector, 512, 1, stream) != 1) {
        restore_error (fs);
        return 0;
    }

    return 1;
}

static int
analyse_boot_sector (msdos_filesystem *fs) {
    fs->sector_size = fs->boot_sector.sector_size[0]
                      + fs->boot_sector.sector_size[1] * 256;
    fs->fat_table_count = fs->boot_sector.fats;
    fs->root_dir_entry_count = fs->boot_sector.dir_entries[0]
                               + fs->boot_sector.dir_entries[1] * 256;
    fs->fat_offset = fs->boot_sector.reserved * fs->sector_size;

    if (fs->fs_type == FAT16) {
        fs->fat_size = fs->boot_sector.fat_length * fs->sector_size;
        fs->fat_entry_count = fs->fat_size / 2;

        fs->root_offset = fs->fat_offset + fs->fat_size * fs->fat_table_count;
    }
    
    if (fs->fs_type == FAT32) {
        fs->fat_size = fs->boot_sector.fat32_fat_length * fs->sector_size;
        fs->fat_entry_count = fs->fat_size / 4;

        fs->info_sector_offset
            = fs->boot_sector.fat32_info_sector * fs->sector_size;
    }

    fs->cluster_offset = fs->fat_offset
                            + fs->fat_size * fs->fat_table_count
                            - 2 * fs->cluster_size;

    return 1;
}

static int
read_fat (FILE *stream, msdos_filesystem *fs) {
    if (!read_heading (stream, fs, "FAT")) return 0;

    fs->fats[0] = malloc (fs->fat_size);
    fs->fats[1] = NULL;

    if (fs->fats[0] == NULL) {
        trimlib_error (fs, "trimlib_restore_backup:  not enough memory");
        return 0;
    }

    if (fread (fs->fats[0], fs->fat_size, 1, stream) != 1) {
        restore_error (fs);
        return 0;
    }

    return 1;
}

static int
read_root_dir (FILE *stream, msdos_filesystem *fs, void *buffer) {
    int root_size;

    if (!read_heading (stream, fs, "Root directory")) return 0;

    if (fs->fs_type == FAT16) {
        root_size = fs->root_dir_entry_count * sizeof (directory_entry);
        if (fread (buffer, root_size, 1, stream) != 1) {
            restore_error (fs);
            return 0;
        }
    }

    if (fs->fs_type == FAT32) {
        if (fread (buffer, fs->cluster_size, 1, stream) != 1) {
            restore_error (fs);
            return 0;
        }
    }

    return 1;
}

/*
    What should I do if the write fails?  Just continue blindly?
*/
int
trimlib_restore_backup (FILE *stream, msdos_filesystem *fs) {
    int     i;

    fs->buffer = (char*) malloc (65536);
    if (fs->buffer==NULL) {
        trimlib_error (fs, "trimlib_restore_backup:  not enough memory");
        return 0;
    }

    if (!read_header (stream, fs)) return 0;
    if (!read_partition_size (stream, fs)) return 0;
    if (!read_filesystem_type (stream, fs)) return 0;
    if (!read_boot_sector (stream, fs)) return 0;
    if (!analyse_boot_sector (fs)) return 0;
    if (!read_fat (stream, fs)) return 0;
    if (!read_root_dir (stream, fs, fs->buffer)) return 0;
    if (!count_free_clusters (fs)) return 0;

    if (!trimlib_write_boot_sector (fs)) return 0;
    if (fs->fs_type == FAT32) trimlib_write_info_sector (fs);
    if (!trimlib_write_fat (fs)) /* return 0 */ ;

    if (fs->fs_type == FAT16)
        trimlib_seek_write (fs, fs->buffer, fs->root_offset,
                            fs->root_dir_entry_count * sizeof(directory_entry));

    if (fs->fs_type == FAT32)
        trimlib_write_cluster (fs, fs->buffer, fs->root_cluster);

    sync ();

    fs->sector_count = fs->partition_size / 512;

    fdiskSetConstraint (&fs->device->table[fs->partition].size,
                        fs->sector_count, fs->sector_count, fs->sector_count,
                        1);

    if (fs->device->write_f (fs->device) != 0) {
        if (errno == EBUSY) {
            trimlib_error (fs,
                   "You should reboot your computer before proceeding.");
            /* don't return error - it's more a warning */
        } else {
              trimlib_error (fs, "Eror writing partition table - %s",
                             strerror (errno));
              return 0;
        }
    }

    sync ();

    free (fs->buffer);
    free (fs->fats[0]);
    
    return 1;
}

