/*
    libresize 0.08
    Copyright (C) 1998-1999  Andrew Clausen  <clausen@alphalink.com.au>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can also be contacted at:

    Andrew Clausen
    18 Shaw St
    Ashwood, 3147
    Victoria, Australia

*/

#include "libresize.h"
#include "../libfdisk/libfdisk.h"
#include "fatio.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

int
trimlib_check_mounted (msdos_filesystem *fs) {
    FILE *mtab_fd;
    char buf[128];

    mtab_fd = fopen ("/etc/mtab", "ro");
    if (mtab_fd == NULL) {
        trimlib_error (fs, "unable to open /etc/mtab");
        return 1;
    }

    while (!feof (mtab_fd)) {
        fgets (buf, 127, mtab_fd);
        if (strncmp ((char *) buf, fs->fs_name, strlen (fs->fs_name)) == 0
            && isspace (buf [strlen (fs->fs_name)]) ) {
            trimlib_error (fs, "device is mounted");
            fclose (mtab_fd);
            return 0;
        }
    }

    fclose (mtab_fd);
    return 1;
}

int
trimlib_open_filesystem (msdos_filesystem *fs) {
    if (!trimlib_check_mounted (fs)) return 0;
        
    fs->fs_handle = open (fs->fs_name, O_RDWR);
    if (fs->fs_handle == 0) {
        trimlib_error (fs, "%s", strerror (errno));
        return 0;
    }
    return 1;
}

int
trimlib_seek_read (msdos_filesystem *fs, char *buf, off_t pos,
                   cluster_t count) {
    int status;

    if (buf == NULL) {
        trimlib_error (fs, "trimlib_seek_read: NULL pointer passed");
        return 0;
    }

    status = lseek (fs->fs_handle, pos, SEEK_SET);

    if (status != pos) {
        if (status < 0) {
            trimlib_error (fs, "%s", strerror (errno));
        } else {
            trimlib_error (fs, "error seeking in device");
        }
        return 0;
    }

    status = read (fs->fs_handle, buf, count);

    if (status != count) {
        if (status < 0)
            trimlib_error (fs, "%s", strerror (errno));
        else
            trimlib_error (fs, "read error - short read");

        return 0;
    }

    return 1;
}

int
trimlib_seek_write (msdos_filesystem *fs, char *buf, off_t pos,
                    cluster_t count) {
    int status;

    if (buf == NULL) {
        trimlib_error (fs, "trimlib_seek_write: NULL pointer passed");
        return 0;
    }

    status = lseek (fs->fs_handle, pos, SEEK_SET);

    if (status != pos) {
        if (status < 0)
            trimlib_error (fs, "%s", strerror (errno));
        else
            trimlib_error (fs, "error seeking in device");

        return 0;
    }

#ifndef READ_ONLY
    status = write (fs->fs_handle, buf, count);

    if (status != count) {
        if (status < 0)
            trimlib_error (fs, "%s", strerror (errno));
        else
            trimlib_error (fs, "write error - short write");

        return 0;
    }
    sync ();
#endif

    return 1;
}

int
trimlib_read_fat (msdos_filesystem *fs) {
    int i;

    fs->new_fat = (void*) malloc (fs->fat_size);

    for (i=0; i<2; i++) {
        fs->fats[i] = (void*) malloc (fs->fat_size);
        if (fs->fats[i] == NULL) {
            trimlib_error (fs, "not enough memory to load FAT");
            return 0;
        }

        if (!trimlib_seek_read (fs, (void *) fs->fats[i],
                                fs->fat_offset + i*fs->fat_size, fs->fat_size))
            return 0;

        if ( *((unsigned char*) fs->fats[i]) != fs->boot_sector.media) {
            trimlib_error (fs, "FAT %d has invalid signature %#.2x",
                           i + 1, (int) * (char*) (fs->fats[i]));
            return 0;
        }
    }

    fs->free_cluster_count = 0;
    fs->bad_cluster_count = 0;      /* every cluster is a BAD cluster :-) */
    fs->used_cluster_count = 0;
    for (i=2; i < fs->cluster_count+2; i++) {
        if (trimlib_is_available (fs, i)) fs->free_cluster_count++;
        if (trimlib_is_bad (fs, i)) fs->bad_cluster_count++;
        if (!trimlib_is_empty (fs, i)) fs->used_cluster_count++;
    }

    return 1;
}

/*
 - compares the two FATs (one's a backup that should match)
 - skips first entry - its just a signature (already checked above)

  NOTE: checks for cross-linking are done in count.c
*/
int
trimlib_check_fat (msdos_filesystem *fs) {
    if (fs->fats[1] == NULL) return 1;

    if (memcmp ((char*) fs->fats[0] + 4, (char*) fs->fats[1] + 4,
                 fs->fat_size - 4)
            != 0) {
        trimlib_error (fs, "FAT tables do not match");
        return 0;
    }

    return 1;
}

void
trimlib_set_fat (msdos_filesystem *fs, cluster_t cluster, cluster_t value) {
    if (cluster > fs->cluster_count+2) {
        trimlib_error (fs, "set_fat: cluster %ld outside filesystem", cluster);
        exit (1);
    }

    switch (fs->fs_type) {
        case FAT16:
            ((unsigned short *) fs->fats[0]) [cluster]
                = ((unsigned short *) fs->fats[1]) [cluster]
                = value;
            break;

        case FAT32:
            ((unsigned int *) fs->fats[0]) [cluster]
                = ((unsigned int *) fs->fats[1]) [cluster]
                = value;
            break;
    }
}

cluster_t 
trimlib_get_fat (msdos_filesystem *fs, cluster_t cluster) {
    if (cluster > fs->cluster_count+2) {
        trimlib_error (fs, "get_fat: cluster %ld outside filesystem", cluster);
        exit (1);
    }

    switch (fs->fs_type) {
        case FAT16: return ((unsigned short *) fs->fats[0]) [cluster];
        case FAT32: return ((unsigned int *) fs->fats[0]) [cluster];
    }
}

int
trimlib_read_clusters (msdos_filesystem *fs, char *buf, cluster_t cluster,
                       long count) {
    off_t   seek_pos;

    if (cluster + count > fs->cluster_count+2) {
        trimlib_error (fs,
                   "trimlib_read_clusters: clusters %ld-%ld outside filesystem",
                       cluster, cluster + count);
        return 0;
    }

    seek_pos = (off_t) fs->cluster_offset;
    seek_pos += ((off_t) cluster) * fs->cluster_size;
    return trimlib_seek_read (fs, buf, seek_pos, count * fs->cluster_size);
}

int
trimlib_read_cluster (msdos_filesystem *fs, char *buf, cluster_t cluster) {
    return trimlib_read_clusters (fs, buf, cluster, 1);
}

int
trimlib_write_clusters (msdos_filesystem *fs, char *buf, cluster_t cluster,
                        long count) {
    off_t   seek_pos;

    if (cluster + count > fs->cluster_count+2) {
        trimlib_error (fs,
                  "trimlib_write_clusters: clusters %ld-%ld outside filesystem",
                       cluster, cluster + count);
        return 0;
    }

    seek_pos = (off_t) fs->cluster_offset;
    seek_pos += ((off_t) cluster) * fs->cluster_size;
    return trimlib_seek_write (fs, buf, seek_pos, count * fs->cluster_size);
}

int
trimlib_write_cluster (msdos_filesystem *fs, char *buf, cluster_t cluster) {
    return trimlib_write_clusters (fs, buf, cluster, 1);
}

int
trimlib_write_fat (msdos_filesystem *fs) {
    int i;

    if (!trimlib_check_fat (fs)) return 0;

    for (i=0; i < fs->fat_table_count; i++) {
        if (!trimlib_seek_write (fs, (char *) fs->fats[0],
                                 fs->fat_offset + i * fs->fat_size,
                                 fs->fat_size))
            return 0;
    }

    return 1;
}

cluster_t
get_free_cluster (msdos_filesystem *fs) {
    cluster_t   i;
    cluster_t   cluster;

    for (i=0; i < fs->cluster_count; i++) {
        cluster = (i + fs->last_free_cluster - 2) % fs->cluster_count + 2;
        if (trimlib_is_available (fs, cluster)) {
            fs->last_free_cluster = cluster;
            return cluster;
        }
    }

    trimlib_error (fs, "get_free_cluster: no free clusters\n");
    return 0;
}

/*
    returns true if <cluster> is marked as bad
*/
int
trimlib_is_bad (msdos_filesystem *fs, cluster_t cluster) {
    cluster_t   code;
    code = trimlib_get_fat (fs, cluster);

    switch (fs->fs_type) {
        case FAT16:
        if (code == 0xfff7) return 1;
        break;

        case FAT32:
        if (code == 0xffffff7) return 1;
        break;
    }
    
    return 0;
}

/*
    returns true if <cluster> represents an EOF marker
*/
int
trimlib_is_eof (msdos_filesystem *fs, cluster_t cluster) {
    switch (fs->fs_type) {
        case FAT16:
        if (cluster >= 0xfff7) return 1;
        break;

        case FAT32:
        if (cluster >= 0xffffff7) return 1;
        break;
    }

    return 0;
}

/*
    returns true if <cluster> is empty.  Note that this includes bad clusters.
*/
int
trimlib_is_empty (msdos_filesystem *fs, cluster_t cluster) {
    cluster_t code;

    code = trimlib_get_fat (fs, cluster);
    if (trimlib_is_bad (fs, cluster)) return 1;
    return (code == 0);
}

/*
    returns true if <cluster> is available.
*/
int
trimlib_is_available (msdos_filesystem *fs, cluster_t cluster) {
    return trimlib_get_fat (fs, cluster) == 0;
}

/*
    marks <cluster> as the last cluster in the chain
*/
void
trimlib_set_eof (msdos_filesystem *fs, cluster_t cluster) {
    switch (fs->fs_type) {
        case FAT16:
        trimlib_set_fat (fs, cluster, 0xfff8);
        break;
        
        case FAT32:
        trimlib_set_fat (fs, cluster, 0xffffff8);
        break;
    }
}

/*
    marks <cluster> as unused/free/available
*/
void
trimlib_set_avail (msdos_filesystem *fs, cluster_t cluster) {
    trimlib_set_fat (fs, cluster, 0);
}

void
trimlib_clear_new_fat (msdos_filesystem *fs) {
    memset (fs->new_fat, 0, fs->fat_size);
    trimlib_set_new_fat(fs, 0, 0x0fffffff);
    trimlib_set_new_fat(fs, 1, 0x0fffffff);
    ((char *) fs->new_fat) [0] = fs->boot_sector.media;
}

void
trimlib_set_new_fat_avail (msdos_filesystem *fs, cluster_t cluster) {
    trimlib_set_new_fat (fs, cluster, 0);
}

void
trimlib_set_new_eof (msdos_filesystem *fs, cluster_t cluster) {
    switch (fs->fs_type) {
        case FAT16:
        trimlib_set_new_fat (fs, cluster, 0xfff8);
        break;
        
        case FAT32:
        trimlib_set_new_fat (fs, cluster, 0xffffff8);
        break;
    }
}

void
trimlib_set_new_fat (msdos_filesystem *fs, cluster_t cluster, cluster_t value) {
    switch (fs->fs_type) {
        case FAT16:
        ((unsigned short *) fs->new_fat) [cluster] = value;
        break;
        
        case FAT32:
        ((unsigned int *) fs->new_fat) [cluster] = value;
        break;
    }
}

cluster_t
trimlib_get_new_fat (msdos_filesystem *fs, cluster_t cluster) {
    switch (fs->fs_type) {
        case FAT16: return ((unsigned short *) fs->new_fat) [cluster];
        case FAT32: return ((unsigned int *) fs->new_fat) [cluster];
    }
}

void
trimlib_use_new_fat (msdos_filesystem *fs) {
    memcpy (fs->fats[0], fs->new_fat, fs->fat_size);
    memcpy (fs->fats[1], fs->new_fat, fs->fat_size);
}
