/*
    libresize 0.08
    Copyright (C) 1998-1999  Andrew Clausen  <clausen@alphalink.com.au>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can also be contacted at:

    Andrew Clausen
    18 Shaw St
    Ashwood, 3147
    Victoria, Australia

*/

#include "libresize.h"
#include "traverse.h"
#include "count.h"
#include "fatio.h"
#include "../libfdisk/libfdisk.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <stdarg.h>

#ifdef USE_NEWT
#include <newt.h>
#endif

#define MIN(a, b) (a<b) ? a : b;
#define MAX(a, b) (a>b) ? a : b;

char *fs_types[] = {
    "FAT12",
    "FAT16",
    "FAT32",
    NULL
};

void
trimlib_print_message (msdos_filesystem *fs, char *message) {
#ifdef USE_NEWT
    newtWinMessage ("Error", "Ok", message);
#else
    fprintf (stderr, message);
#endif
}

/*
    Generates an error message
*/
void
trimlib_error (msdos_filesystem *fs, char *format, ...) {
    char tmp [256];
    va_list args;

    va_start (args, format);
    vsnprintf (tmp, 256, format, args);
    va_end (args);

    snprintf (fs->error_msg, 256, "libresize: %s: %s\n", fs->fs_name, tmp);
    trimlib_print_message (fs, fs->error_msg);
}

/*
    Generates an error message, without displaying the device name
*/
void
trimlib_custom_error (msdos_filesystem *fs, char *format, ...) {
    char tmp [256];
    va_list args;

    va_start (args, format);
    vsnprintf (tmp, 256, format, args);
    va_end (args);

    snprintf (fs->error_msg, 256, "libresize: %s\n", tmp);
    trimlib_print_message (fs, fs->error_msg);
}

static int
check_fdisk_error (msdos_filesystem *fs, int status) {
    if (status < 0) {
        trimlib_custom_error (fs, "%s: %s", fs->device_name, strerror (errno));
        return 0;
    } else if (status > 0) {
        trimlib_custom_error (fs, "%s: error opening device", fs->device_name);
        return 0;
    }
    return 1;
}

int
trimlib_check_null (void *ptr, msdos_filesystem *fs, char *error_msg) {
    if (ptr == NULL) {
        trimlib_error (fs, error_msg);
        return 0;
    }
    return 1;
}

/*
    opens the partition (using libfdisk), and gets some info out of it.
*/
int
trimlib_open_partition_table (msdos_filesystem *fs) {
    int temp;
    int dummy;
    int status;
    
    status = fdiskOpenDevice (fs->device_name, 1, &fs->device);
    if (!check_fdisk_error(fs, status)) return 0;

    status = fdiskReadPartitions (fs->device);
    if (!check_fdisk_error(fs, status)) return 0;

    fdiskGetCurrentConstraint (&fs->device->table[fs->partition].start, &temp);
    fs->partition_start = ((long long) temp) * 512;

    fdiskGetCurrentConstraint (&fs->device->table[fs->partition].size, &temp);
    fs->partition_size = ((long long) temp) * 512;

    fdiskGetMinMaxConstraint (&fs->device->table[fs->partition].size, &dummy,
                              &temp);
    /* need to go and complain to Red Hat */
//    fs->max_partition_size = temp * 512;

    return 1;
}

/*
 init_filesystem()

 - reads in the boot sector/partition info., and tries to make some sense of it
 - reads in FATs
 - returns 1 on success, 0 on failure

parameters:
  filesystem    an empty structure to fill.
  device_name   name of the device. eg /dev/hda
  partition     partition number of file system. eg 3 for /dev/hda3

*/
int
trimlib_init_filesystem (msdos_filesystem *fs, char *device_name,
                 int partition) {

    if (!trimlib_check_null (fs, NULL, "NULL filesystem structure passed"))
        return 0;
    if (!trimlib_check_null (device_name, fs, "NULL device name passed"))
        return 0;

    fs->device_name = device_name;
    fs->partition   = partition;
    fs->fs_name = malloc (strlen(device_name) + 3);
    sprintf(fs->fs_name, "%s%d", device_name, partition);

    if (!trimlib_open_filesystem (fs)) return 0;
    if (!trimlib_open_partition_table (fs)) return 0;

    return 1;
}

int
trimlib_allocate_buffers (msdos_filesystem *fs) {
    for (fs->buffer_size = 4;
         fs->buffer_size*fs->cluster_size <= MAX_BUFFER_SIZE;
         fs->buffer_size *= 2)
    {
        fs->buffer = malloc (fs->buffer_size * fs->cluster_size);
        if (fs->buffer == (char*) NULL) break;
        free (fs->buffer);
    }

    fs->buffer_size /= 4;
    if (fs->buffer_size < 4) {
        trimlib_error (fs, "not enough memory to resize filesystem");
        return 0;
    }
    fs->buffer = malloc (fs->buffer_size * fs->cluster_size);
    fs->buffer2 = malloc (fs->buffer_size * fs->cluster_size);

    fs->last_free_cluster = 2;
    fs->fat_remap = malloc (fs->fat_entry_count * sizeof (long));

    fs->fat_flag_map = malloc (fs->fat_entry_count);

    return 1;
}

/*
    prepares a filesystem to "do stuff".  Performs checks, analyses bits
    & pieces.  NOTE:  this isn't in trimlib_init_filesystem (anymore), because
    some applications (i.e. restoring backups) most drop the requirement of
    having a healthy filesystem.
*/
int
trimlib_prepare_filesystem (msdos_filesystem *fs) {
    if (!trimlib_check_null (fs, fs, "trimlib_print_info: NULL passed"))
        return 0;

    if (!trimlib_init_boot_sector (fs)) return 0;
    if (fs->fs_type == FAT32)
        if (!trimlib_read_info_sector (fs)) return 0;
    if (!trimlib_allocate_buffers (fs)) return 0;
    if (!trimlib_read_fat (fs)) return 0;
    if (!trimlib_check_fat (fs)) return 0;
    if (!trimlib_flag_clusters (fs)) return 0;

    return 1;
}

/*
    unprepare - for lack of a better word  (cleanup might have some
    negative conotations)  see prepare if you're confused.
*/
void
trimlib_unprepare_filesystem (msdos_filesystem *fs) {
    if (!trimlib_check_null (fs, NULL, "NULL filesystem structure passed"))
        return;

    free (fs->fats[0]);
    free (fs->fats[1]);
    free (fs->new_fat);
    free (fs->fat_remap);

    free (fs->buffer);
    free (fs->buffer2);
}

void
trimlib_close_filesystem (msdos_filesystem *fs) {
    int status;

    if (!trimlib_check_null (fs, NULL, "NULL filesystem structure passed"))
        return;

    status = fdiskCloseDevice (fs->device);
    check_fdisk_error (fs, status);

    close (fs->fs_handle);
    free (fs->fs_name);

    sync ();
}

/*
    allocates where all the clusters will be moved to.  Clusters before
    cut_point remain in the same position, however cluster that are
    part of a directory are moved regardless (this is a mechanism to prevent
    data loss)

    (cut_point is the first cluster that won't occur in the new fs)
*/
int
trimlib_allocate_remap (msdos_filesystem *fs, cluster_t cut_point) {
    cluster_t   cluster_no;
    cluster_t   new_cluster;

    for (cluster_no=2; cluster_no<cut_point; cluster_no++) {
        if (fs->fat_flag_map[cluster_no] == FAT_FLAGS_DIRECTORY) {
            if (fs->fs_type == FAT32 && cluster_no == fs->root_cluster) {
                // don't move this, because it has to remain in the first 64k
                // clusters
                fs->fat_remap [cluster_no] = cluster_no;
                continue;
            }

            new_cluster = get_free_cluster (fs);
            if (new_cluster == 0 || new_cluster >= cut_point) return 0;
            trimlib_set_eof (fs, new_cluster);   // mark as used
            
            fs->fat_remap [cluster_no] = new_cluster;
#ifdef DEBUG
            printf ("[%d,%d...]->%d...\n", cluster_no,
                    trimlib_get_fat (fs, cluster_no), new_cluster);
#endif
        } else {
            fs->fat_remap [cluster_no] = cluster_no;
        }
    }

    for (cluster_no=cut_point; cluster_no < fs->cluster_count+2; cluster_no++) {
        if (!trimlib_is_empty (fs, cluster_no)) {
            new_cluster = get_free_cluster (fs);
            if (new_cluster == 0 || new_cluster >= cut_point) return 0;
            trimlib_set_eof (fs, new_cluster);   // mark as used
            
            fs->fat_remap [cluster_no] = new_cluster;

#ifdef DEBUG
            printf ("%d->%d\n", cluster_no, new_cluster);
#endif
        }
    }

    return 1;
}

/*
    Advances *walk until a cluster that hold information for a file is found.
    If *walk reaches final_cluster first, then it returns 0, otherwise 1.
*/
static int
find_file_cluster (msdos_filesystem *fs, cluster_t *walk,
                   cluster_t final_cluster) {

    while (fs->fat_flag_map[*walk] != FAT_FLAGS_FILE
           && (*walk) <= final_cluster)
        (*walk)++;
    
    return (*walk) <= final_cluster;
}

/*
    calculates which cluster (that has been read in) will be the last to fit
    in the write buffer (i.e. fs->buffer2).  From this, calculates the size
    of the read_size and write_size

    It assumes that as the "read" cluster increases (represented by walk),
    the remapped "write" cluster also increases.  The algorithm for remapping
    should ensure this.
*/
static void
calc_run_size (msdos_filesystem *fs, cluster_t first_cluster,
               cluster_t final_cluster, long *read_size, long *write_size) {
    cluster_t   walk;
    cluster_t   first_write_cluster;
    cluster_t   last_write_cluster;
    cluster_t   ok_last_write_cluster;
    cluster_t   ok_last_read_cluster;

    first_write_cluster = fs->fat_remap [first_cluster];
    ok_last_read_cluster = first_cluster;
    ok_last_write_cluster = first_write_cluster;

    for (walk=first_cluster; walk<=final_cluster; walk++) {
        if (fs->fat_flag_map[walk] != FAT_FLAGS_FILE) continue;

        last_write_cluster = fs->fat_remap[walk];
        if (last_write_cluster - first_write_cluster + 1 >= fs->buffer_size)
            break;

        ok_last_read_cluster = walk;
        ok_last_write_cluster = last_write_cluster;
    }

    *read_size = ok_last_read_cluster - first_cluster + 1;
    *write_size = ok_last_write_cluster - first_write_cluster + 1;
}

static void
overlay_changed_clusters (msdos_filesystem *fs, cluster_t first_read_cluster,
                          cluster_t first_write_cluster,
                          long run_size, char *read_buffer) {
    long    read_cluster;
    long    write_cluster;

    for (read_cluster = first_read_cluster;
         read_cluster < first_read_cluster + run_size;
         read_cluster++) {
        
        if (fs->fat_flag_map[read_cluster] != FAT_FLAGS_FILE) continue;

        write_cluster = fs->fat_remap[read_cluster];
        memcpy (&fs->buffer2[(write_cluster - first_write_cluster)
                                 * fs->cluster_size],
                &read_buffer[(read_cluster - first_read_cluster)
                                 * fs->cluster_size],
                fs->cluster_size);
    }
}

/*
    reads as many clusters as possible.  In the case of error, just
    skips over...
*/
static void
tolerant_read_clusters (msdos_filesystem *fs, char *buffer,
                        cluster_t first_cluster, long count) {
    cluster_t   cluster;
    
    if (!trimlib_read_clusters (fs, buffer, first_cluster, count)) {
        for (cluster=first_cluster; cluster < first_cluster+count; cluster++)
            trimlib_read_cluster (fs, &buffer[cluster * fs->cluster_size],
                                  cluster);
    }
}

/*
    writes as many clusters as possible.  In the case of error, just
    skips over...
*/
static void
tolerant_write_clusters (msdos_filesystem *fs, char *buffer,
                         cluster_t first_cluster, long count) {
    cluster_t   cluster;
    
    if (!trimlib_write_clusters (fs, buffer, first_cluster, count)) {
        for (cluster=first_cluster; cluster < first_cluster+count; cluster++)
            trimlib_write_cluster (fs, &buffer[cluster * fs->cluster_size],
                                   cluster);
    }
}

/*
    copies all clusters that need to be moved (determined by allocate_remap),
    THAT ARE FILES.  Directory entries will need to be modified, so they're
    done separately.  (in trimlib_construct_dir_tree)
*/
int
trimlib_place_clusters (msdos_filesystem *fs,
                        cluster_t start_cluster, long cluster_count) {
    cluster_t   read_cluster;
    cluster_t   final_cluster;
    long        run_size;
    cluster_t   write_cluster;
    long        write_size;
    char       *run_ptr;

    final_cluster = start_cluster + cluster_count - 1;

    for (read_cluster = start_cluster; read_cluster <= final_cluster;
         read_cluster += run_size) {

        if (!find_file_cluster (fs, &read_cluster, final_cluster)) return 1;

#ifdef DEBUG
        printf ("   copy cluster %ld...\n", read_cluster);
#endif

        run_ptr = fs->buffer + (read_cluster - start_cluster)*fs->cluster_size;
        write_cluster = fs->fat_remap[read_cluster];

        calc_run_size (fs, read_cluster, final_cluster, &run_size, &write_size);

        // overlay the clusters to be written
        tolerant_read_clusters (fs, fs->buffer2, write_cluster, write_size);
        overlay_changed_clusters (fs, read_cluster, write_cluster, run_size,
                                  run_ptr);
        tolerant_write_clusters (fs, fs->buffer2, write_cluster, write_size);
    }
    
    return 1;
}

/*
    copy all clusters >= <start_cluster> to a new place on the partition,
    less than <start_cluster>.  Only copies files, not directories.
*/
int
trimlib_copy_clusters (msdos_filesystem *fs, cluster_t start_cluster) {
    long        i;
    cluster_t   cluster;
    long        transfer_size;

    cluster = start_cluster;
    while (cluster < fs->cluster_count + 2) {
        while (cluster < fs->cluster_count + 2
               && fs->fat_flag_map[cluster] != FAT_FLAGS_FILE)
            cluster++;

        if (cluster >= fs->cluster_count + 2) return 1;

        transfer_size = MIN (fs->buffer_size, fs->cluster_count + 2 - cluster);
        
        if (!trimlib_read_clusters (fs, fs->buffer, cluster, transfer_size)) {
            /* read one by one - avoiding known bad clusters */
            for (i=cluster; i<cluster+transfer_size; i++)
                if (!trimlib_is_empty (fs, i))
                    if (!trimlib_read_cluster
                            (fs, &fs->buffer [i * fs->cluster_size], i))
                        return 0;
        }

        if (!trimlib_place_clusters (fs, cluster, transfer_size))
            return 0;
        
        cluster += fs->buffer_size;
    }

    return 1;
}

/*
    Recursively duplicates a directory tree, modified for the new file
    locations.
*/
int
trimlib_construct_directory (trimlib_traverse_info *traverse_info) {
    msdos_filesystem       *fs = traverse_info->fs;
    trimlib_traverse_info  *sub_dir_info;
    directory_entry        *dir_entry;
    directory_entry        *new_dir;
    directory_entry        *new_entry;
    int                     entry_num;

    new_dir = (directory_entry*) malloc (fs->cluster_size);
    if (!trimlib_check_null (new_dir, fs,
                "Not enough memory for directory buffer")) return 0;

    new_entry = new_dir;
    entry_num = 0;
    while ( (dir_entry = trimlib_get_next (traverse_info)) ) {
        *new_entry = *dir_entry;

        if (trimlib_has_first_cluster (fs, dir_entry))
            trimlib_file_set_first_cluster (fs, new_entry,
                 fs->fat_remap[trimlib_file_get_first_cluster (fs, dir_entry)]);

        if (trimlib_is_directory (dir_entry) && dir_entry->name[0] != '.') {
            sub_dir_info
                = trimlib_traverse_directory (traverse_info, dir_entry);
            if (!trimlib_construct_directory (sub_dir_info)) return 0;
        }
 
        new_entry++;
        entry_num++;
        if (entry_num == entries_per_buffer (traverse_info)) {
            if (!trimlib_write_cluster (fs, (void*) new_dir,
                     fs->fat_remap[traverse_info->this_cluster])) return 0;
            entry_num = 0;
            new_entry = new_dir;
        }
    }

// pad the rest of the cluster with NULLs
    if (entry_num != 0) {
        while (entry_num < entries_per_buffer (traverse_info)) {
            memset ((void*) new_entry, 0, sizeof (directory_entry));
            new_entry++;
            entry_num++;
        }

        if (!trimlib_write_cluster (fs, (void*) new_dir,
                   fs->fat_remap[traverse_info->this_cluster])) {
            trimlib_traversal_complete (traverse_info);
            return 0;
        }
    }

    free (new_dir);
    trimlib_traversal_complete (traverse_info);
    return 1;
}

/*
    Constructs the new directory tree for legacy (FAT16) filesystems.
*/
int
trimlib_construct_legacy_root (msdos_filesystem *fs) {
    trimlib_traverse_info  *traverse_info;
    trimlib_traverse_info  *subdir_info;
    directory_entry        *dir_entry;
    directory_entry        *new_root_dir;
    directory_entry        *new_entry;
    int                     entry_num;
    int                     root_sector_num;

    new_root_dir = (directory_entry*) 
                   malloc (fs->root_dir_entry_count * sizeof (directory_entry));
    if (!trimlib_check_null (new_root_dir, fs,
                "Not enough memory for root directory buffer")) return 0;
    
    traverse_info = trimlib_traversal_begin (fs, FAT_ROOT, "\\");

    entry_num = 0;
    new_entry = new_root_dir;
    root_sector_num = 0;
    while ( (dir_entry = trimlib_get_next (traverse_info)) ) {
        *new_entry = *dir_entry;

        if (trimlib_has_first_cluster (fs, dir_entry))
            trimlib_file_set_first_cluster (fs, new_entry,
                 fs->fat_remap[trimlib_file_get_first_cluster (fs, new_entry)]);

        if (trimlib_is_directory (dir_entry)) {
            subdir_info = trimlib_traverse_directory (traverse_info, dir_entry);
            if (!trimlib_construct_directory (subdir_info)) return 0;
        }
            
        entry_num++;
        new_entry++;

        if (entry_num == entries_per_buffer (traverse_info)) {
            if (!trimlib_seek_write (fs, (void*) new_root_dir,
                fs->root_offset + root_sector_num * fs->sector_size,
                fs->sector_size))
                return 0;
            entry_num = 0;
            new_entry = new_root_dir;
            root_sector_num++;
        }
    }

// This ugly code pads the rest of the root dir with NULLs
    if (entry_num != 0) {
        while (entry_num < entries_per_buffer (traverse_info)) {
            memset ((void*) new_entry, 0, sizeof (directory_entry));
            new_entry++;
            entry_num++;
        }

        if (!trimlib_seek_write (fs, (void*) new_root_dir,
                fs->root_offset + root_sector_num * fs->sector_size,
                fs->sector_size))
            return 0;

        root_sector_num++;
    }

    memset ((void*) new_root_dir, 0, fs->sector_size);

    while (root_sector_num
             <  fs->root_dir_entry_count / entries_per_buffer (traverse_info)) {
        trimlib_seek_write (fs, (void*) new_root_dir,
            fs->root_offset + root_sector_num * fs->sector_size,
            fs->sector_size);
        root_sector_num++;
    }

    free (new_root_dir);
    trimlib_traversal_complete (traverse_info);
    return 1;
}

/*
    Constructs the new directory tree for new (FAT32) filesystems.
*/
int
trimlib_construct_root (msdos_filesystem *fs) {
    trimlib_traverse_info  *traverse_info;

    traverse_info = trimlib_traversal_begin (fs, fs->root_cluster, "\\");
    trimlib_construct_directory (traverse_info);
    return 1;
}

/*
    Constructs the new directory tree to match the new file locations.
*/
int
trimlib_construct_dir_tree (msdos_filesystem *fs) {
    switch (fs->fs_type) {
        case FAT16: return trimlib_construct_legacy_root (fs);
        case FAT32: return trimlib_construct_root (fs);
    }
}

/*
    constructs the new fat for the resized filesystem.
*/
int
trimlib_construct_new_fat (msdos_filesystem *fs) {
    cluster_t   old_cluster;
    cluster_t   new_cluster;
    cluster_t   old_next_cluster;

    trimlib_clear_new_fat (fs);
    
    for (old_cluster=2; old_cluster < fs->cluster_count+2; old_cluster++) {
        if (fs->fat_flag_map[old_cluster] != FAT_FLAGS_FREE) {
            new_cluster = fs->fat_remap[old_cluster];
            if (new_cluster != old_cluster)
                trimlib_set_new_fat_avail (fs, old_cluster);

            old_next_cluster = trimlib_get_fat (fs, old_cluster);
            if (trimlib_is_eof (fs, old_next_cluster))
                trimlib_set_new_eof (fs, new_cluster);
            else
                trimlib_set_new_fat (fs, new_cluster,
                                     fs->fat_remap[old_next_cluster]);
        }
    }

#ifdef DEBUG
    for (old_cluster=2; old_cluster < fs->cluster_count+2; old_cluster++) {
        if (trimlib_get_fat (fs, old_cluster)
              != trimlib_get_new_fat (fs, old_cluster)) {
            printf ("FAT entry for cluster %d changed:  %d->%d\n",
                     old_cluster, trimlib_get_fat (fs, old_cluster),
                     trimlib_get_new_fat (fs, old_cluster));
        } else if (trimlib_get_fat (fs, old_cluster) != 0) {
            printf ("%d:  %d\n",
                     old_cluster, trimlib_get_fat (fs, old_cluster));
        }
    }
#endif

    trimlib_use_new_fat (fs);

    return 1;
}

/*
    returns the number of clusters for a given filesystem type
*/
cluster_t
trimlib_min_cluster_count (int fs_type) {
    switch (fs_type) {
        case FAT16: return 0xff4;
        case FAT32: return 0xfff4;
    }
}

cluster_t
trimlib_max_cluster_count (int fs_type) {
    switch (fs_type) {
        case FAT16: return 0xfff5;
        case FAT32: return 0xfffffff5;
    }
}

// returns cylinder size in sectors
int
trimlib_cylinder_size (msdos_filesystem *fs) {
    return fs->device->geom.heads * fs->device->geom.sectors;
}

/*
    rounds up sector_count such that if the fs was resized, it would end on
    a cylinder boundary, ensuring that at least sector_count sectors would be
    included
*/
int
trimlib_round_upto_cylinder (msdos_filesystem *fs, int sector_count) {
    int cylinder_size;      // cylinder size in sectors
    int cylinder_count;
    int new_sector_count;
    int start_offset;       // the number of sectors after the cylinder boundary
                            // where the partition starts

    cylinder_size = trimlib_cylinder_size (fs);
    start_offset = fs->device->table[fs->partition].start.current
                       % cylinder_size;

    cylinder_count = sector_count / cylinder_size;

    new_sector_count = cylinder_count * cylinder_size - start_offset;
    while (new_sector_count < sector_count) {
        cylinder_count++;
        new_sector_count = cylinder_count * cylinder_size - start_offset;
    }

    return new_sector_count;
}

/*
    calculates the minimum size of a partition, in physical sectors (as opposed
    to M$'s bloody "logical" sectors)  The filesystem must be in the prepared
    state (see trimlib_prepare_filesystem)
*/
cluster_t
trimlib_calc_min_size (msdos_filesystem *fs) {
    cluster_t   min_cluster_count;
    cluster_t   min_sector_count;

    min_cluster_count = fs->used_cluster_count + fs->bad_cluster_count
                            + fs->total_dir_clusters;

/* make sure we keep above M$'s braindead limitations */
    min_cluster_count = MAX (min_cluster_count,
                             trimlib_min_cluster_count (fs->fs_type));

    min_sector_count = min_cluster_count * (fs->cluster_size / 512)
                           + fs->cluster_offset / 512;

/* Round up to nearest cylinder.  Lots of programs depend on this (eg. lilo) */
    return min_sector_count;
}

/*
    TODO: see if libfdisk can calculate the maximum size of the partition
          for us.
*/
cluster_t
trimlib_calc_max_size (msdos_filesystem *fs) {
    int max_cluster_count;
    int max_sector_count;

    max_cluster_count = MIN (fs->fat_entry_count - 2,
                             trimlib_max_cluster_count (fs->fs_type));

    max_sector_count = max_cluster_count * (fs->cluster_size / 512)
                           + fs->cluster_offset / 512;

    return max_sector_count;
}

/*
  new_partition_size is in bytes

  - checks boundaries before starting
  - copies all data beyond new_cluster_count behind the frontier
  - assumes sector size (physical, not logical) is always 512 bytes (can
    someone confirm this?)
*/
int
trimlib_resize_filesystem (msdos_filesystem *fs, long long new_partition_size) {
    int     new_data_size;
    int     new_sector_count;
    int     new_cluster_count;
    int     i;

    if (!trimlib_prepare_filesystem (fs)) return 0;

    new_sector_count = new_partition_size / 512;

    if (new_sector_count < trimlib_calc_min_size (fs)) {
        trimlib_error (fs, "Minimum filesystem size is %d sectors",
                       trimlib_calc_min_size (fs)); 
        return 0;
    }

    if (new_sector_count > trimlib_calc_max_size (fs)) {
        trimlib_error (fs, "Maximum filesystem size is %d sectors",
                       trimlib_calc_max_size (fs));
        return 0;
    }

    new_partition_size
        = trimlib_round_upto_cylinder (fs, new_sector_count) * 512;

    new_data_size = new_partition_size - fs->cluster_offset;
    new_cluster_count = new_data_size / fs->cluster_size;

#ifdef DEBUG
    printf ("changing filesystem size:  %d->%d clusters\n",
            fs->cluster_count, new_cluster_count);
    printf ("copying clusters...\n");
#endif

    if (!trimlib_allocate_remap (fs, new_cluster_count)) return 0;
    if (!trimlib_copy_clusters (fs, new_cluster_count)) return 0;

#ifdef DEBUG
    printf ("writing new directory tree...\n");
#endif

    if (!trimlib_construct_dir_tree (fs)) return 0;

#ifdef DEBUG
    printf ("writing new FAT...\n");
#endif

    if (!trimlib_construct_new_fat (fs)) return 0;
    if (!trimlib_write_fat (fs)) return 0;

    fs->sector_count = new_sector_count;
    fs->cluster_count = new_cluster_count;
    fs->free_cluster_count = fs->cluster_count - fs->used_cluster_count
                                - fs->bad_cluster_count;

    fs->boot_sector.sectors[0] = 0;
    fs->boot_sector.sectors[1] = 0;
    fs->boot_sector.total_sect = new_sector_count;

#ifdef DEBUG
    printf ("writing new boot sector...\n");
#endif

    if (!trimlib_write_boot_sector (fs)) return 0;
    if (fs->fs_type == FAT32)
        trimlib_write_info_sector (fs);  // doesn't matter if this fails - its
                                         // pretty useless!

    sync();

    fdiskSetConstraint (&fs->device->table[fs->partition].size,
                        fs->sector_count, fs->sector_count,
                        fs->sector_count, 1);

#ifndef READ_ONLY
    if (fs->device->write_f (fs->device) != 0) {
         if (errno == EBUSY) {
             trimlib_error (fs,
                          "You should reboot your computer before proceeding.");
         } else {
             trimlib_error (fs, "%s: %s.  Partition table write failed.\n",
                            fs->device_name, strerror (errno));
             return 0;
         }
    }
#endif

    sync ();

    trimlib_unprepare_filesystem (fs);
    return 1;
}

void
trimlib_print_info (msdos_filesystem *fs) {
    if (!trimlib_prepare_filesystem (fs)) return;

    printf ("filesystem type:         %s\n", fs_types[fs->fs_type]);

    printf ("cluster size:            %ld bytes\n", fs->cluster_size);
    printf ("no. of clusters:         %ld\n", fs->cluster_count);
    printf ("no. used:                %ld\n", fs->used_cluster_count);
    printf ("no. free:                %ld\n", fs->free_cluster_count);

    printf ("FAT offset:              0x%lx\n", fs->fat_offset);
    if (fs->fs_type == FAT32)
        printf ("Root cluster:            %d\n", fs->root_cluster);
    else
        printf ("root offset:             0x%lx\n", fs->root_offset);
    printf ("cluster offset:          0x%lx\n", fs->cluster_offset);

    printf ("partition start sector:  %d\n", fs->partition_start/512);
    printf ("partition size:          %d sectors\n", fs->partition_size/512);
    printf ("filesystem size:         %d\n", fs->sector_count);

    printf ("minimum partition size:  %d sectors\n",
            trimlib_round_upto_cylinder (fs, trimlib_calc_min_size (fs)));
    printf ("maximum partition size: ~%d sectors\n",
            trimlib_round_upto_cylinder (fs, trimlib_calc_max_size (fs)));

    trimlib_unprepare_filesystem (fs);
}
